# -*- coding: utf-8 -*-
""" This module describes the signals that are used on this package.

There are 3 kinds of signal:

**Position signal**

Will be sent after each event come to notify current binlog's position.

    def subscriber((log_file, log_pos):
        pass

**Binlog signal**

Will be sent for each binlog event, it could be Update/Write/Delete event.

    def subscriber(event, stream):
        pass

**Rows signals** : *the most important ones*

They will be sent on each binlog event, after convert event into a dict.

    def subscriber(table_name, rows, meta):
        pass

Where table_name if a string in form of schema.table. It is also blinker's
sender, so if you want to subscribe for one table only, just specify it via
sender argument of *connect* method.

There are 3 signals: rows_inserted, rows_updated, rows_deleted
"""
import blinker

__author__ = 'tarzan'


_signals = blinker.Namespace()


# position signal
binlog_position_signal = _signals.signal(
    'binlog_position'
)
""":type: blinker.NamedSignal"""

# fired on each RowsEvent come
# def subscriber(event, stream)
binlog_signal = _signals.signal(
    'event',
    doc='fired on each RowsEvent come',
)
""":type: blinker.NamedSignal"""
on_binlog = binlog_signal.connect

# def subscriber(table_name, rows, meta)
rows_inserted = _signals.signal(
    'rows_inserted',
    doc='fired on each WriteRowsEvent come, after convert data into dict',
)
""":type: blinker.NamedSignal"""
on_rows_inserted = rows_inserted.connect

# def subscriber(table_name, rows, meta)
rows_updated = _signals.signal(
    'rows_updated',
    doc='fired on each UpdateRowsEvent come, after convert data into dict',
)
""":type: blinker.NamedSignal"""
on_rows_updated = rows_updated.connect

# def subscriber(table_name, rows, meta)
rows_deleted = _signals.signal(
    'rows_deleted',
    doc='fired on each DeleteRowsEvent come, after convert data into dict',
)
""":type: blinker.NamedSignal"""
on_rows_deleted = rows_deleted.connect



# coding=utf-8
import logging

from pymysqlreplication import row_event

from mysqlbinlog2blinker import signals

__author__ = 'Tarzan'
_logger = logging.getLogger(__name__)


def _get_updated_values(before_values, after_values):
    """ Get updated values from 2 dicts of values

    Args:
        before_values (dict): values before update
        after_values (dict): values after update

    Returns:
        dict: a diff dict with key is field key, value is tuple of
              (before_value, after_value)
    """
    assert before_values.keys() == after_values.keys()
    return dict([(k, [before_values[k], after_values[k]])
                 for k in before_values.keys()
                 if before_values[k] != after_values[k]])


def _convert_write_row(row):
    """ Convert a row for write/delete event """
    return row


def _convert_update_row(row):
    """ Convert a row for update event

    Args:
        row (dict): event row data
    """
    after_values = row['after_values']  # type: dict
    before_values = row['before_values']  # type: dict
    values = after_values
    return {
        'values': values,
        'updated_values': _get_updated_values(before_values, after_values)
    }


def _rows_event_to_dict(e, stream):
    """ Convert RowsEvent to a dict

    Args:
        e (pymysqlreplication.row_event.RowsEvent): the event
        stream (pymysqlreplication.BinLogStreamReader):
            the stream that yields event

    Returns:
        dict: event's data as a dict
    """
    pk_cols = e.primary_key if isinstance(e.primary_key, (list, tuple)) \
        else (e.primary_key, )

    if isinstance(e, row_event.UpdateRowsEvent):
        sig = signals.rows_updated
        action = 'update'
        row_converter = _convert_update_row
    elif isinstance(e, row_event.WriteRowsEvent):
        sig = signals.rows_inserted
        action = 'insert'
        row_converter = _convert_write_row
    elif isinstance(e, row_event.DeleteRowsEvent):
        sig = signals.rows_deleted
        action = 'delete'
        row_converter = _convert_write_row
    else:
        assert False, 'Invalid binlog event'

    meta = {
        'time': e.timestamp,
        'log_pos': stream.log_pos,
        'log_file': stream.log_file,
        'schema': e.schema,
        'table': e.table,
        'action': action,
    }
    rows = list(map(row_converter, e.rows))
    for row in rows:
        row['keys'] = {k: row['values'][k] for k in pk_cols}
    return rows, meta


@signals.on_binlog
def on_binlog(event, stream):
    """ Process on a binlog event

    1. Convert event instance into a dict
    2. Send corresponding schema/table/signals

    Args:
        event (pymysqlreplication.row_event.RowsEvent): the event
    """
    rows, meta = _rows_event_to_dict(event, stream)

    table_name = '%s.%s' % (meta['schema'], meta['table'])

    if meta['action'] == 'insert':
        sig = signals.rows_inserted
    elif meta['action'] == 'update':
        sig = signals.rows_updated
    elif meta['action'] == 'delete':
        sig = signals.rows_deleted
    else:
        raise RuntimeError('Invalid action "%s"' % meta['action'])

    sig.send(table_name, rows=rows, meta=meta)


# -*- coding: utf-8 -*-
import logging
import os
import sys
import threading
from mysqlbinlog2blinker import signals

__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


class BaseBinlogPosMemory(object):
    """ This class will receive binlog position signal from mysqlbinlog2blinker
    and store it into a persistent storage.
    """
    def start(self):
        """ Start listening the signal and monitoring position changing
        """
        raise NotImplementedError()

    def stop(self):
        """ Finish listening the signal and monitoring position changing
        """
        raise NotImplementedError()

    @property
    def log_pos(self):
        """ Return current or last saved position """
        raise NotImplementedError()

    @property
    def log_file(self):
        """ Return current or last saved binlog file """
        raise NotImplementedError()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class FileBasedBinlogPosMemory(BaseBinlogPosMemory):
    """ This class store binlog position into file.

    It will subscribe to mysqlbinlog2blinker.signals.binlog_pos_signal and store the
    position into memory.

    A thread will run and store this position into a file each interval.
    The position in file will be read the the memory start.
    """
    def __init__(self, pos_filename, interval=2):
        """ Create instance of FileBasedBinlogPosMemory

        Args:
            pos_filename (str|None): position storage file. None will makes
                *mysqlbinlog2blinker.binlog.pos* at current working dir
            interval (float): the interval in second
        """
        if not pos_filename:
            pos_filename = os.path.join(os.getcwd(),
                                        'mysqlbinlog2blinker.binlog.pos')
        self.pos_storage_filename = pos_filename
        assert self.pos_storage_filename
        self.interval = interval

        self._log_file = None
        self._log_pos = None
        self._pos_changed = False

        self.save_log_pos_thread_stop_flag = threading.Event()
        self.save_log_pos_thread = \
            threading.Thread(target=self._save_log_pos_thread_runner)
        self.save_log_pos_thread.daemon = True

    def _save_log_pos_thread_runner(self):
        """ Save the position into file after every second
        :return:
        """
        _logger.debug('Start log pos saving thread')
        while not self.save_log_pos_thread_stop_flag.wait(self.interval):
            self._save_file_and_pos()
        _logger.debug('Finish log pos saving thread')

    @property
    def log_pos(self):
        return self._log_pos

    @property
    def log_file(self):
        return self._log_file

    def set_binlog_pos(self, log_file, log_pos):
        if (log_file != self._log_file) or (log_pos != self._log_pos):
            self._log_file, self._log_pos = log_file, log_pos
            self._pos_changed = True

    def on_binlog_pos_signal(self, pos):
        self.set_binlog_pos(*pos)

    def start(self):
        _logger.debug('Start binlog position memory at %s'
                      % self.pos_storage_filename)
        self._read_file_and_pos()
        signals.binlog_position_signal.connect(self.on_binlog_pos_signal)
        self.save_log_pos_thread_stop_flag.clear()
        self.save_log_pos_thread.start()

    def stop(self):
        signals.binlog_position_signal.disconnect(self.on_binlog_pos_signal)
        self.save_log_pos_thread_stop_flag.set()
        _logger.debug('Finish binlog position memory at %s'
                      % self.pos_storage_filename)
        # flushing current position into file
        self._save_file_and_pos()

    def _save_file_and_pos(self):
        """ Save current position into file
        """
        if not self._pos_changed:
            return
        with open(self.pos_storage_filename, 'w+') as f:
            _pos = '%s:%s' % (self._log_file, self._log_pos)
            _logger.debug('Saving position %s to file %s'
                          % (_pos, self.pos_storage_filename))
            f.write(_pos)
            self._pos_changed = False

    def _read_file_and_pos(self):
        """ Read last position from file, store as current position
        """
        try:
            with open(self.pos_storage_filename, 'r+') as f:
                _pos = f.read()
                _logger.debug('Got position "%s" from file %s'
                              % (_pos, self.pos_storage_filename))
                if not _pos:
                    return
                log_file, log_pos = _pos.split(':')
                try:
                    log_file = str(log_file)
                    log_pos = int(log_pos)
                except (ValueError, TypeError) as e:
                    _logger.critical('Can not read position: %s' % e)
                    sys.exit(13)
                self._log_file = log_file
                self._log_pos = log_pos
        except IOError as e:
            _logger.error(e)


# -*- coding: utf-8 -*-
"""MySQL binlog to Blinker

This project listen on mysql binlog (as a slave) and send the binlog event to
blinker signal. It can remember last binlog position and work as a MySQL slave.
"""
import logging

__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


# The package name, which is also the "UNIX name" for the project.
package = 'mysqlbinlog2blinker'
project = "MySQL binlog to Blinker signal"
project_no_spaces = project.replace(' ', '')
version = '0.1'
description = 'This project listen on mysql binlog (as a slave) and send the ' \
              'binlog event to blinker signal. It can remember last binlog ' \
              'position and work as a MySQL slave.'
authors = ['Tarzan', ]
authors_string = ', '.join(authors)
emails = ['hoc3010@gmail.com', ]
license = 'MIT'
copyright = '2016 ' + authors_string
url = 'https://github.com/tarzanjw/python-mysql-binlog-to-blinker'


# -*- coding: utf-8 -*-
import logging
import random

import pymysqlreplication
from pymysqlreplication import row_event
from mysqlbinlog2blinker import (
    _subscribers,
    binlog_pos_memory as _bpm,
    signals,
)


__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


def start_publishing(mysql_settings, **kwargs):
    """Start publishing MySQL row-based binlog events to blinker signals

    Args:
        mysql_settings (dict): information to connect to mysql via pymysql
        **kwargs: The additional kwargs will be passed to
        :py:class:`pymysqlreplication.BinLogStreamReader`.
    """
    _logger.info('Start publishing from %s with:\n%s'
                 % (mysql_settings, kwargs))

    kwargs.setdefault('server_id', random.randint(1000000000, 4294967295))
    kwargs.setdefault('freeze_schema', True)

    # connect to binlog stream
    stream = pymysqlreplication.BinLogStreamReader(
        mysql_settings,
        only_events=[row_event.DeleteRowsEvent,
                     row_event.UpdateRowsEvent,
                     row_event.WriteRowsEvent],
        **kwargs
    )
    """:type list[RowsEvent]"""

    for event in stream:
        # ignore non row events
        if not isinstance(event, row_event.RowsEvent):
            continue

        _logger.debug('Send binlog signal "%s@%s.%s"' % (
            event.__class__.__name__,
            event.schema,
            event.table
        ))
        signals.binlog_signal.send(event, stream=stream)
        signals.binlog_position_signal.send((stream.log_file, stream.log_pos))


def start_replication(mysql_settings,
                      binlog_pos_memory=(None, 2),
                      **kwargs):
    """ Start replication on server specified by *mysql_settings*

    Args:
         mysql_settings (dict): mysql settings that is used to connect to
                                mysql via pymysql
        binlog_pos_memory (_bpm.BaseBinlogPosMemory):
            Binlog Position Memory, it should be an instance of subclass of
            :py:class:`_bpm.BaseBinlogPosMemory`.

            If a tuple (str, float) is passed, it will be initialize parameters
            for default :py:class:`_bpm.FileBasedBinlogPosMemory`. It the file-
            name is None, it will be *`cwd`\mysqlbinlog2blinker.binlog.pos*
        **kwargs: any arguments that are accepted by
            :py:class:`pymysqlreplication.BinLogStreamReader`'s constructor
    """
    if not isinstance(binlog_pos_memory, _bpm.BaseBinlogPosMemory):
        if not isinstance(binlog_pos_memory, (tuple, list)):
            raise ValueError('Invalid binlog position memory: %s'
                             % binlog_pos_memory)
        binlog_pos_memory = _bpm.FileBasedBinlogPosMemory(*binlog_pos_memory)

    mysql_settings.setdefault('connect_timeout', 5)
    kwargs.setdefault('blocking', True)
    kwargs.setdefault('resume_stream', True)

    with binlog_pos_memory:
        kwargs.setdefault('log_file', binlog_pos_memory.log_file)
        kwargs.setdefault('log_pos', binlog_pos_memory.log_pos)
        _logger.info('Start replication from %s with:\n%s'
                     % (mysql_settings, kwargs))

        start_publishing(mysql_settings, **kwargs)